#include "app.h"

#define FRAND_BASE (10000)
#define ATTRACT_WAIT (1)

enum{
	ENEMY_01 = 0,
	ENEMY_02,
	ENEMY_03,
	ENEMY_04,
	ENEMY_05,
	ENEMY_06,
	ENEMY_07,
	ENEMY_08,
};

static void TSKgctrlExit(TSK *tskp);
static void TSKgctrlDraw(TSK *tskp);

static int TSKgctrlAttractChk(TSK *tskp);
static void TSKgctrlSndClear(TSK *tskp);
static void TSKgctrlTskClear(TSK *tskp);
static int TSKgctrlStage(TSK *tskp);
static void initEnemyTable(void);
static void shuffleEnemyTable(void);
static void bonusCalc(void);
static void nextStage(void);

static int req_bgm;
static int attract_wait;
static int attract_stg = 0;
static int rank;
static TSK *tsk_title;
static TSK *tsk_ranking;

static FUNC enemy_func[] = {
								&TSKzako01,
								&TSKzako02,
								&TSKzako03,
								&TSKzako04,
								&TSKzako05,
								&TSKzako06,
								&TSKzako07,
								&TSKzako08,
							};
static int enemy_table[64];
static int enemy_tblmax;
static int enemy_tblnum;

static int STAGE_LOG = 0;

void TSKgctrl(TSK *tskp)
{
	int i;
	int ret;

	switch(tskp->step){
		case	0:
			tskp->exit_adr = &TSKgctrlExit;
			tskp->draw_adr = &TSKgctrlDraw;
			tsk_title = TSKsetBtm(&TSKtitle, GROUP_SYSTEM);
			TSKsetBtm(&TSKstar, GROUP_BG);
			scr_vel[X] = 0;
			scr_vel[Y] = 0;
			stage = 0;
			play_stg = 0;
			playtime = 0;
			attract_mode = 0;
			game_mode = 0;
			pause_flag = 0;
#if ATTRACT_WAIT
			tskp->wait = 60 * 15;
#else
			tskp->wait = 60 * 1;
#endif
			tskp->step++;
			break;
		case	1:
			if(SDL_PCB_CoinIsFreeplay()){
				if(pcb_trgs & PAD_A){
					TSKclear(tsk_title);
					tskp->step = 3;
					break;
				}
			}
			if(SDL_PCB_CoinGetOneCredit() >= 1){
				tskp->step++;
				break;
			}
			if(SDL_PCB_CoinIsFreeplay() || SDL_PCB_CoinIsNoneCredit()){
				if(tskp->wait){
					tskp->wait--;
				}else{
					attract_wait = 60 * 60;
					tskp->step = 100;
				}
			}
			break;
		case	2:
			if(pcb_trgs & PAD_A){
				TSKclear(tsk_title);
				tskp->step = 3;
				break;
			}
			break;
		case	3:
			SDL_PCB_CoinDecCredit();
			SDL_PCB_TestIncPlayCnt();
			pause_flag = 1;
			stage = 0;
			play_stg = 0;
			req_bgm = -1;
			score = 0;
			left = 2;
			enemy_base = 0;
			game_mode = 1;
			initEnemyTable();
			TSKsetBtm(&TSKship, GROUP_SHIP);
			TSKsetTop(&TSKsysinf, GROUP_SYSTEM);
			TSKsetBtm(&TSKalert, GROUP_SYSTEM);
			tskp->step++;
		case	4:
			TSKsetBtm(&TSKnoise, GROUP_SYSTEM);
			noise_flag = 0;
			SDL_PCB_SndPlaySe(SND_STGCLR);
			if((rand() % 100) > 50){
				scr_vel[X] = frand();
				scr_vel[Y] = 0;
			}else{
				scr_vel[X] = 0;
				scr_vel[Y] = frand();
			}
			ret = (stage / 8) % 3;
			switch(ret){
				case	0:
					if(req_bgm != SND_BGM01) SDL_PCB_SndPlayMusic(SND_BGM01);
					break;
				case	1:
					if(req_bgm != SND_BGM02) SDL_PCB_SndPlayMusic(SND_BGM02);
					break;
				case	2:
					if(req_bgm != SND_BGM03) SDL_PCB_SndPlayMusic(SND_BGM03);
					break;
			}
			shuffleEnemyTable();
			req_bgm = ret + SND_BGM01;
			tskp->wait = 90;
			tskp->step++;
			break;
		case	5:
			if(noise_flag == 1){
				enemy_max = 4 + enemy_base;
				enemy_cnt = 8 + stage * 2;
				enemy_born = 0;
				enemy_now = 0;
				enemy_dest = 0;;
				if(enemy_max > 16) enemy_max = 16;
				if(enemy_cnt > 32) enemy_cnt = 32;
				TSKclearGroup(tskp, GROUP_SSHOT);
				TSKclearGroup(tskp, GROUP_ESHOT);
				noise_flag = 2;
			}
			if(tskp->wait){
				tskp->wait--;
				break;
			}
			TSKsetBtm(&TSKzone, GROUP_SYSTEM);
			stagetime = playtime;
			if(tsk_ship) tsk_ship->step = 1;
			tskp->mode = 0;
			tskp->step++;
			break;
		case	6:
			playtime++;
			if(game_mode == 1){
				ret = TSKgctrlStage(tskp);
				if(ret){
					tskp->step++;
				}
			}else if(game_mode == 2){
				TSKsetBtm(&TSKgameover, GROUP_FRONT);
				SDL_PCB_SndSetFadeMusic(1, 90);
				tskp->wait = 180;
				tskp->step = 8;
			}
			break;
		case	7:
			if(tsk_ship){
				tsk_ship->step = 3;
				tsk_ship->wait = 15;
			}
			play_stg++;
			stagetime = playtime - stagetime;
			nextStage();
			bonusCalc();
			TSKsetTop(&TSKcleartime, GROUP_FRONT);
			tskp->step = 4;
			break;
		case	8:
			if(tskp->wait){
				tskp->wait--;
			}else{
				TSKgctrlSndClear(tskp);
				TSKgctrlTskClear(tskp);
				tskp->wait = 90;
				tskp->step++;
			}
			break;
		case	9:
			if(tskp->wait){
				tskp->wait--;
			}else{
				rank = ranking_set();
				if(rank == -1){
					tskp->step = 0;
				}else{
					tskp->step = 10;
				}
			}
			break;

		case	10:
			initEnemyTable();
			TSKsetBtm(&TSKstar, GROUP_BG);
			req_bgm = -1;
			score = 0;
			left = 2;
			attract_wait = 60 * 10;
			attract_mode = 2;
			TSKsetBtm(&TSKship, GROUP_SHIP);
			TSKsetTop(&TSKsysinf, GROUP_SYSTEM);
			tsk_ranking = TSKsetTop(&TSKranking, GROUP_FRONT);
			if(tsk_ranking){
				tsk_ranking->prog_adr(tsk_ranking);
				tsk_ranking->mode = rank;
			}
			enemy_base = stage;
			for(i = 0; i < stage; i++){
				if(i && !(i % 8)) enemy_base -= 4;
			}
			enemy_tblmax += (stage / 4) * 4;
			tskp->step = 101;
			break;

		case	100:
			initEnemyTable();
			TSKclear(tsk_title);
			req_bgm = -1;
			score = 0;
			left = 2;
			attract_mode = 1;

			///pause_flag = 1;	vCfɃ|[Ŷ͌Ȃ̂ŁAƒfō폜܂Bby N.SAWA

			TSKsetBtm(&TSKship, GROUP_SHIP);
			TSKsetTop(&TSKsysinf, GROUP_SYSTEM);
			if(attract_stg > 23) attract_stg = 0;
			stage = attract_stg;
			enemy_base = stage;
			for(i = 0; i < stage; i++){
				if(i && !(i % 8)) enemy_base -= 4;
			}
			enemy_tblmax += (stage / 4) * 4;
			TSKsetBtm(&TSKattract, GROUP_FRONT);
			tskp->step++;
		case	101:
			if(TSKgctrlAttractChk(tskp)) break;
			TSKsetBtm(&TSKnoise, GROUP_SYSTEM);
			TSKsetBtm(&TSKalert, GROUP_SYSTEM);
			noise_flag = 0;
			SDL_PCB_SndPlaySe(SND_STGCLR);
			if((rand() % 100) > 50){
				scr_vel[X] = frand();
				scr_vel[Y] = 0;
			}else{
				scr_vel[X] = 0;
				scr_vel[Y] = frand();
			}
			ret = (stage / 8) % 3;
			switch(ret){
				case	0:
					if(req_bgm != SND_BGM01) SDL_PCB_SndPlayMusic(SND_BGM01);
					break;
				case	1:
					if(req_bgm != SND_BGM02) SDL_PCB_SndPlayMusic(SND_BGM02);
					break;
				case	2:
					if(req_bgm != SND_BGM03) SDL_PCB_SndPlayMusic(SND_BGM03);
					break;
			}
			shuffleEnemyTable();
			req_bgm = ret + SND_BGM01;
			tskp->wait = 90;
			tskp->step++;
			break;
		case	102:
			if(TSKgctrlAttractChk(tskp)) break;
			if(noise_flag == 1){
				enemy_max = 4 + enemy_base;
				enemy_cnt = 8 + stage * 2;
				enemy_born = 0;
				enemy_now = 0;
				enemy_dest = 0;;
				if(enemy_max > 16) enemy_max = 16;
				if(enemy_cnt > 32) enemy_cnt = 32;
				TSKclearGroup(tskp, GROUP_SSHOT);
				TSKclearGroup(tskp, GROUP_ESHOT);
				noise_flag = 2;
			}
			if(tskp->wait){
				tskp->wait--;
				break;
			}
			TSKsetBtm(&TSKzone, GROUP_SYSTEM);
			stagetime = playtime;
			if(tsk_ship) tsk_ship->step = 1;
			tskp->mode = 0;
			tskp->step++;
			break;
		case	103:
			if(TSKgctrlAttractChk(tskp)) break;
			playtime++;
			ret = TSKgctrlStage(tskp);
			if(ret){
				tskp->step++;
			}
			break;
		case	104:
			if(TSKgctrlAttractChk(tskp)) break;
			if(tsk_ship){
				tsk_ship->step = 3;
				tsk_ship->wait = 15;
			}
			play_stg++;
			stagetime = playtime - stagetime;
			nextStage();
			bonusCalc();
			TSKsetTop(&TSKcleartime, GROUP_FRONT);
			tskp->step = 101;
			break;
		case	105:
			TSKgctrlSndClear(tskp);
			TSKgctrlTskClear(tskp);
			tskp->wait = 90;
			tskp->step++;
			break;
		case	106:
			if(tskp->wait){
				tskp->wait--;
			}else{
				tskp->step = 0;
			}
			break;

		default:
			TSKclear(tskp);
			break;
	}
}

static void TSKgctrlExit(TSK *tskp)
{
}

static void TSKgctrlDraw(TSK *tskp)
{
}

static int TSKgctrlAttractChk(TSK *tskp)
{
	if(SDL_PCB_CoinIsFreeplay()){
		if(pcb_trgs & PAD_A){
			attract_wait = 0;
		}
	}
	if(SDL_PCB_CoinGetOneCredit() >= 1){
		attract_wait = 0;
	}

	if(attract_wait){
#if ATTRACT_WAIT
		attract_wait--;
#endif
		if(attract_mode == 1){
			if(attract_wait == (60 * 17)){
				tsk_ranking = TSKsetTop(&TSKranking, GROUP_FRONT);
				if(tsk_ranking){
					tsk_ranking->prog_adr(tsk_ranking);
					tsk_ranking->mode = -1;
				}
			}
		}
		if(attract_wait == (60 *  2)) SDL_PCB_SndSetFadeMusic(1, 60);
	}else{
		attract_stg = stage;
		tskp->step = 105;
		return 1;
	}

	return 0;
}

void TSKenemyPosSet(TSK *tskp)
{
	tskp->pos[X] = fld(80 + (rand() % 13) * 32 + 16);
	tskp->pos[Y] = fld( 0 + (rand() % 13) * 32 + 16);
}

void TSKpalleteSet(TSK *tskp, int r, int g, int b, int a)
{
	tskp->color.r = r;
	tskp->color.g = g;
	tskp->color.b = b;
	tskp->color.a = a;
	tskp->pal = MapRGBA(r, g, b, a);
}

void TSKareaClamp(TSK *tskp)
{
	     if(fcom(tskp->pos[X], >=, GRP_MAX_X)) fsub_(tskp->pos[X], GRP_SIZE_X);
	else if(fcom(tskp->pos[X], < , GRP_MIN_X)) fadd_(tskp->pos[X], GRP_SIZE_X);
	     if(fcom(tskp->pos[Y], >=, GRP_MAX_Y)) fsub_(tskp->pos[Y], GRP_SIZE_Y);
	else if(fcom(tskp->pos[Y], < , GRP_MIN_Y)) fadd_(tskp->pos[Y], GRP_SIZE_Y);
}

void enemy_destroy(TSK *tskp)
{
	TSK *etskp;

	addScore(tskp->score);
	enemy_now--;
	enemy_dest++;

	if(tsk_ship){
		if(tsk_ship->trg_tsk && tsk_ship->trg_tsk == tskp){
			tsk_ship->trg_tsk = NULL;
		}
	}

	etskp = TSKsetBtm(&TSKscore, GROUP_EFFECT);
	if(etskp){
		etskp->pos[X] = tskp->pos[X];
		etskp->pos[Y] = tskp->pos[Y];
		etskp->cnt = tskp->score;
	}
}

fixed getShipDist(TSK *tskp)
{
	fixed dist;
	fixed px,py;

	px = fsub(tsk_ship->pos[X], tskp->pos[X]);
	py = fsub(tsk_ship->pos[Y], tskp->pos[Y]);	/* \[X͌? pyXW߂Ă̂ŁAYWɏC */
	px = fmul(px, px);	/* \[X͏py,px߂ۂɐΒlĂA */
	py = fmul(py, py);	/* @2悵ĕKɂȂ̂ŁAΒl̏ȗ */
	dist = fsqrt(fadd(px, py));

	return dist;
}

void TBLshuffle(int *tbl, int max)
{
	int i, seed, tmp;

	for(i = max - 1; i > 0; i--){
		seed = rand() % i;
		tmp = tbl[i];
		tbl[i] = tbl[seed];
		tbl[seed] = tmp;
	}
}

fixed frand(void)
{
	return fsub(FRND(seed), fld(0.5));
}

int isTimeUp(void)
{
	if(limit_time) return 0;

	return 1;
}

void addScore(int add_score)
{
	if(!tsk_ship) return;
	if(tsk_ship->step >= 100) return;
	score += add_score;
}

static int TSKgctrlStage(TSK *tskp)
{
	TSK *etskp;
	int seq_mode = 0;
	int ene_num;

	switch(tskp->mode){
		case	0:
			tskp->wait = 60;
			tskp->mode++;
			break;
		case	1:
			if(enemy_now != enemy_max){
				if(tskp->wait){
					tskp->wait--;
				}else{
					tskp->mode++;
				}
			}
			break;
		case	2:
			if(enemy_cnt != enemy_born){
				if(enemy_now < enemy_max){
					ene_num = enemy_table[enemy_tblnum];
					if(STAGE_LOG){
						printf("set enemy-id %d(%d)\n", ene_num, enemy_tblnum);
					}
					etskp = TSKsetBtm(enemy_func[ene_num], GROUP_ENEMY);
					if(etskp){
						enemy_born++;
						enemy_now++;
						tskp->wait = 30;
						tskp->mode = 1;
					}
					enemy_tblnum++;
					enemy_tblnum %= enemy_tblmax;
				}
			}
			if(enemy_cnt <= enemy_dest){
				tskp->mode++;
			}
			break;
		case	3:
			if(tsk_ship->step == 1 || tsk_ship->step == 2){
				seq_mode = 1;
			}
			break;
	}

	return seq_mode;
}

static void TSKgctrlSndClear(TSK *tskp)
{
	SDL_PCB_SndStopMusic();
	SDL_PCB_SndStopSe(SND_STGCLR);
	SDL_PCB_SndStopSe(SND_EXTEND);
	SDL_PCB_SndStopSe(SND_SHIP);
	SDL_PCB_SndStopSe(SND_SDEST);
	SDL_PCB_SndStopSe(SND_REFLECT);
	SDL_PCB_SndStopSe(SND_ENEMY);
	SDL_PCB_SndStopSe(SND_EDEST);
	SDL_PCB_SndPlaySe(SND_STGCLR);
}

static void TSKgctrlTskClear(TSK *tskp)
{
	TSKclearGroup(tskp, GROUP_BG);
	TSKclearGroup(tskp, GROUP_EFFECT);
	TSKclearGroup(tskp, GROUP_ENEMY);
	TSKclearGroup(tskp, GROUP_SSHOT);
	TSKclearGroup(tskp, GROUP_ESHOT);
	TSKclearGroup(tskp, GROUP_SHIP);
	TSKclearGroup(tskp, GROUP_SYSTEM);
	TSKclearGroup(tskp, GROUP_FRONT);
	TSKsetBtm(&TSKnoise, GROUP_SYSTEM);
}

static void initEnemyTable(void)
{
	int i;

	for(i = 0; i < 4; i++){
		enemy_table[i] = ENEMY_01;
	}
	for(i = 4; i < 8; i++){
		enemy_table[i] = ENEMY_02;
	}
	for(i = 8; i < 12; i++){
		enemy_table[i] = ENEMY_03;
	}
	for(i = 12; i < 16; i++){
		enemy_table[i] = ENEMY_04;
	}
	for(i = 16; i < 20; i++){
		enemy_table[i] = ENEMY_05;
	}
	for(i = 20; i < 24; i++){
		enemy_table[i] = ENEMY_06;
	}
	for(i = 24; i < 28; i++){
		enemy_table[i] = ENEMY_07;
	}
	for(i = 28; i < 32; i++){
		enemy_table[i] = ENEMY_08;
	}
	for(i = 32; i < 64; i += 8){
		enemy_table[i+0] = ENEMY_01;
		enemy_table[i+1] = ENEMY_02;
		enemy_table[i+2] = ENEMY_03;
		enemy_table[i+3] = ENEMY_04;
		enemy_table[i+4] = ENEMY_05;
		enemy_table[i+5] = ENEMY_06;
		enemy_table[i+6] = ENEMY_07;
		enemy_table[i+7] = ENEMY_08;
	}

	enemy_tblmax = 4;
}

static void shuffleEnemyTable(void)
{
	enemy_tblnum = 0;
	enemy_tblmax += (stage && !(stage % 4) ? 4 : 0);
	if(enemy_tblmax > 64) enemy_tblmax = 64;
	TBLshuffle(enemy_table, enemy_tblmax);

	if(STAGE_LOG){
		int i;
		printf("next stage %d\n", stage);
		printf("enemy table max %d(%d)\n", enemy_tblmax, enemy_tblnum);
		printf("enemy set table list\n");
		for(i = 0; i < 64; i++){
			if(i && !(i % 16)) printf("\n");
			printf("%02d,", enemy_table[i]);
		}
		printf("\n");
	}
}

static void bonusCalc(void)
{
	if(!dead_flag){
		if(limit_time > (60 * 25)){
			time_bonus = 50000;
		}else if(limit_time > (60 * 20)){
			time_bonus = 10000;
		}else if(limit_time > (60 * 15)){
			time_bonus = 5000;
		}else if(limit_time > (60 * 10)){
			time_bonus = 1000;
		}else if(limit_time > (60 * 5)){
			time_bonus = 500;
		}else if(limit_time && limit_time < 5){
			time_bonus = 100000;
		}else{
			time_bonus = 0;
		}
	}else{
		time_bonus = 0;
	}
}

static void nextStage(void)
{
	int i, ret;

	if(!dead_flag){
		if(enemy_cnt < 16){
			skip_stage_3 = ((enemy_cnt * 45) / 60) * 60;
			skip_stage_2 = ((enemy_cnt * 55) / 60) * 60;
		}else{
			skip_stage_3 = ((enemy_cnt * 30) / 60) * 60;
			skip_stage_2 = ((enemy_cnt * 40) / 60) * 60;
		}
	}else{
		skip_stage_3 = 0;
		skip_stage_2 = 0;
	}
	skip_stage = skip_stage_2;

	if(stagetime < skip_stage_3)	  ret = 3;
	else if(stagetime < skip_stage_2) ret = 2;
	else							  ret = 1;
	for(i = 0; i < ret; i++){
		stage++;
		enemy_base++;
		if(stage && !(stage % 8)) enemy_base -= 4;
	}
}
