#include "app.h"

#define RANKING_FNAME	"ranking.dat"

static void TSKrankingExit(TSK *tskp);
static void TSKrankingDraw(TSK *tskp);
static void setRankingGrp(TSK* tskp);
static void TSKrankingSetMode(TSK *tskp);

static char *rank_str[] = {
							" 1ST",
							" 2ND",
							" 3RD",
							" 4TH",
							" 5TH",
							" 6TH",
							" 7TH",
							" 8TH",
							" 9TH",
							"10TH",
						};

void TSKranking(TSK *tskp)
{
	switch(tskp->step){
		case	0:
			tskp->exit_adr = &TSKrankingExit;
			tskp->draw_adr = &TSKrankingDraw;
			TSKrankingSetMode(tskp);
			tskp->wait = 60 * 15;
			tskp->cnt = 0;
			break;
		case	1:
			tskp->pos[X] = fadd(fmul(tskp->trg[X], tskp->rate), fmul(tskp->vel[X], fsub(fild(1), tskp->rate)));	/* ݈ʒu = ڕWʒu*rate + Oʒu*(1-rate) */
			tskp->pos[Y] = fadd(fmul(tskp->trg[Y], tskp->rate), fmul(tskp->vel[Y], fsub(fild(1), tskp->rate)));
			fadd_(tskp->rate, fadd(fidiv(fsub(fild(1), tskp->rate), 5), fld(0.01/**/)));
			if(ficom(tskp->rate, >, 1)) tskp->rate = fild(1);
			tskp->cnt++;
			if(tskp->wait){
				tskp->wait--;
			}else{
				tskp->step = -1;
			}
			break;
		case	2:
			tskp->scl[X] = fadd(fmul(tskp->trg[X], tskp->rate), fmul(tskp->vel[X], fsub(fild(1), tskp->rate)));	/* ]TCY = ڕWTCY*rate + OTCY*(1-rate) */
			tskp->scl[Y] = fadd(fmul(tskp->trg[Y], tskp->rate), fmul(tskp->vel[Y], fsub(fild(1), tskp->rate)));
			fadd_(tskp->rate, fld(1.0 / 15.0));
			if(ficom(tskp->rate, >, 1)) tskp->rate = fild(1);
			tskp->cnt++;
			if(tskp->wait){
				tskp->wait--;
			}else{
				tskp->step = -1;
			}
			break;
		default:
			TSKclear(tskp);
			break;
	}
}

static void TSKrankingExit(TSK *tskp)
{
}

static void clipSDLmask(fixed px, fixed py, fixed w, fixed h)
{
	int x0;
	int y0;
	int x1;
	int y1;

	fsub_(px, fidiv(w, 2));
	fsub_(py, fidiv(h, 2));

	px = MapX(px);
	py = MapY(py);
	 w = MapX( w);
	 h = MapY( h);

	fadd_(w, fld(0.5)); /*  */
	fadd_(h, fld(0.5)); /*  */

	x0 = fist(px);
	y0 = fist(py);
	x1 = x0 + fist(w);
	y1 = y0 + fist(h);

	if(x0 <      0) x0 =      0;
	if(y0 <      0) y0 =      0;
	if(x1 > DISP_X) x1 = DISP_X;
	if(y1 > DISP_Y) y1 = DISP_Y;

	render.left   = x0;
	render.top    = y0;
	render.right  = x1;
	render.bottom = y1;

	if(now & 1) {
		render_rectangle_dark(&render, 0, 0, DISP_X, DISP_Y, -1, 0);
	}
}

static void TSKrankingDraw(TSK *tskp)
{
	clipSDLmask(tskp->pos[X], tskp->pos[Y], tskp->scl[X], tskp->scl[Y]);
	/*NbsO*/
	setRankingGrp(tskp);
	/*NbsO*/
	render.left   =      0;
	render.top    =      0;
	render.right  = DISP_X;
	render.bottom = DISP_Y;
}

void ranking_init(void)
{
	int i;

	for(i = 0; i < RANK_MAX; i++){
		ranking[i].score = 1000 * (RANK_MAX - i);
		ranking[i].stage = RANK_MAX - i - 1;
		ranking[i].playtime = 60 * 60 * 60 * 100 - 1;
		ranking[i].avgtime = 60 * 30;
	}
}

int ranking_save(void)
{
	int result = 0;
	FILEACC fa;

	if(pceFileCreate(RANKING_FNAME, sizeof ranking) != 0) {
		fprintf(stderr, "save : ranking data create error.\n");
	} else {
		if(pceFileOpen(&fa, RANKING_FNAME, FOMD_WR) != 0) {
			fprintf(stderr, "save : ranking data open error.\n");
		} else {
			if(pceFileWriteSct(&fa, ranking, 0, sizeof ranking) != sizeof ranking) {
				fprintf(stderr, "save : ranking data write error.\n");
			} else {
				result = 1;
			}
			pceFileClose(&fa);
		}
	}

	return result;
}

int ranking_load(void)
{
	int result = 0;
	FILEACC fa;

	if(pceFileOpen(&fa, RANKING_FNAME, FOMD_RD) != 0) {
		fprintf(stderr, "load : ranking data open error.\n");
	} else {
		if(pceFileReadSct(&fa, ranking, 0, sizeof ranking) != sizeof ranking) {
			fprintf(stderr, "load : ranking data read error.\n");
		} else {
			result = 1;
		}
		pceFileClose(&fa);
	}

	return result;
}

int ranking_set(void)
{
	int i, j;

	for(i = 0; i < RANK_MAX; i++){
		if(score < ranking[i].score) continue;
		if(score == ranking[i].score){
			if(ranking[i].stage >= stage) continue;
			if(ranking[i].playtime >= playtime) continue;
			if(ranking[i].avgtime >= (playtime / stage)) continue;
		}
		for(j = RANK_MAX - 1; j > i; j--){
			ranking[j] = ranking[j-1];
		}
		ranking[i].score    = score;
		ranking[i].stage    = stage;
		ranking[i].playtime = playtime;
		ranking[i].avgtime  = playtime / stage;
		return i;
	}

	return -1;
}

static void setRankingGrp(TSK* tskp)
{
	char str[256];
	int i;
	int time, m, s;
	fixed x, y;

	y = fadd(fsub(tskp->pos[Y], fimul(ASC_SIZE_Y, 6)), fidiv(ASC_SIZE_Y, 2));

	strcpy(str, "- TOP 10 RANKING -");
	x = fadd(fsub(tskp->pos[X], fidiv(getWidthASCII(str), 2)), fidiv(ASC_SIZE_X, 2));
	drawASCII(str, x, y);
	fadd_(y, ASC_SIZE_Y);

	strcpy(str, "RANK SCORE STAGE TIME ");
	x = fadd(fsub(tskp->pos[X], fidiv(getWidthASCII(str), 2)), fidiv(ASC_SIZE_X, 2));
	drawASCII(str, x, y);
	fadd_(y, ASC_SIZE_Y);

	strcpy(str, "10TH 99999999 99 99:99");
	x = fadd(fsub(tskp->pos[X], fidiv(getWidthASCII(str), 2)), fidiv(ASC_SIZE_X, 2));

	for(i = 0; i < RANK_MAX; i++){
		time = ranking[i].playtime;
		m    = (time / (60 * 60)) % 100;
		s    = (time /  60      ) %  60;
		sprintf(str, "%s %08d %02d %02d:%02d", rank_str[i], ranking[i].score, ranking[i].stage + 1, m, s);
		if((i != tskp->mode) || !(now & 0x20)) {
			drawASCII(str, x, y);
		}
		fadd_(y, ASC_SIZE_Y);
	}
}

static void TSKrankingSetMode(TSK *tskp)
{
	int ret = rand() % 4;

	switch(ret){
		case	0:
			tskp->pos[X] = fild( 320);		/* ݈ʒu */
			tskp->pos[Y] = fild(-240);
			tskp->vel[X] = tskp->pos[X];	/* Oʒu */
			tskp->vel[Y] = tskp->pos[Y];
			tskp->trg[X] = fild(320);		/* ڕWʒu */
			tskp->trg[Y] = fild(240);
			tskp->scl[X] = fild(450);		/* ]TCY */
			tskp->scl[Y] = fild(400);
			tskp->rate = 0;
			tskp->step = 1;
			break;
		case	1:
			tskp->pos[X] = fild(320);		/* Œʒu */
			tskp->pos[Y] = fild(240);
			tskp->scl[X] = fild(450);		/* ݃TCY */
			tskp->scl[Y] = fild(200);
			tskp->vel[X] = tskp->scl[X];	/* OTCY */
			tskp->vel[Y] = tskp->scl[Y];
			tskp->trg[X] = fild(450);		/* ڕWTCY */
			tskp->trg[Y] = fild(400);
			tskp->rate = 0;
			tskp->step = 2;
			break;
		case	2:
			tskp->pos[X] = fild(320);		/* Œʒu */
			tskp->pos[Y] = fild(240);
			tskp->scl[X] = fild(225);		/* ݃TCY */
			tskp->scl[Y] = fild(400);
			tskp->vel[X] = tskp->scl[X];	/* OTCY */
			tskp->vel[Y] = tskp->scl[Y];
			tskp->trg[X] = fild(450);		/* ڕWTCY */
			tskp->trg[Y] = fild(400);
			tskp->rate = 0;
			tskp->step = 2;
			break;
		case	3:
			tskp->pos[X] = fild(320);		/* Œʒu */
			tskp->pos[Y] = fild(240);
			tskp->scl[X] = fild(225);		/* ݃TCY */
			tskp->scl[Y] = fild(200);
			tskp->vel[X] = tskp->scl[X];	/* OTCY */
			tskp->vel[Y] = tskp->scl[Y];
			tskp->trg[X] = fild(450);		/* ڕWTCY */
			tskp->trg[Y] = fild(400);
			tskp->rate = 0;
			tskp->step = 2;
			break;
		default:
			DIE();
	}
}
