#include "app.h"

#define ACC_MAX  fld(6.00)
#define ACC_RATE fld(0.08)
#define DIR_BLANK (4)

#define EXTEND_1ST (50000)
#define EXTEND_2ND (100000)
#define EXTEND_3RD (200000)

#define TIME_OFF (0)

#if !TIME_OFF
#define STG_LIMIT_TIME (60 * 30)
#else
#define STG_LIMIT_TIME (60 * 0)
#endif

static void TSKshipInt(TSK *tskp);
static void TSKshipExit(TSK *tskp);
static void TSKshipDraw(TSK *tskp);
static void TSKshipAccCtrl(TSK *tskp);
static void TSKshipShotCtrl(TSK *tskp);
static void TSKshipExtend(TSK *tskp);
static void TSKshipTimeCtrl(TSK *tskp);
static void TSKshipAttract(TSK *tskp);
static void ship_destroy(TSK *tskp);

static VERTEX vertex[] = {
							{ fld( +0.0), fld(-12.0) },
							{ fld(-12.0), fld(+12.0) },
							{ fld( +0.0), fld( +3.0) },
							{ fld(+12.0), fld(+12.0) },
						};

static fixed ship_vel[][XYZ] = {
								{       0 ,       0 ,                 0 },	/* 0x00 (none) */
								{ fild( 2),       0 ,              PI_2 },	/* 0x01 PAD_RI */
								{ fild(-2),       0 , (fixed)-(int)PI_2 },	/* 0x02 PAD_LF */
								{       0 ,       0 ,                 0 },	/* 0x03 (none) */
								{       0 , fild( 2),                PI },	/* 0x04 PAD_DN */
								{       0 ,       0 ,                 0 },	/* 0x05 (none) */
								{       0 ,       0 ,                 0 },	/* 0x06 (none) */
								{       0 ,       0 ,                 0 },	/* 0x07 (none) */
								{       0 , fild(-2),                 0 },	/* 0x08 PAD_UP */
								{       0 ,       0 ,                 0 },	/* 0x09 (none) */
								{       0 ,       0 ,                 0 },	/* 0x0A (none) */
								{       0 ,       0 ,                 0 },	/* 0x0B (none) */
								{       0 ,       0 ,                 0 },	/* 0x0C (none) */
								{       0 ,       0 ,                 0 },	/* 0x0D (none) */
								{       0 ,       0 ,                 0 },	/* 0x0E (none) */
								{       0 ,       0 ,                 0 },	/* 0x0F (none) */
							};

static int ship_pads;
static int ship_dirm;
static int extend_step;
static int extend_score;
static int shot_wait;

static int time_off = TIME_OFF;

void TSKship(TSK *tskp)
{
	if(!attract_mode){
		ship_pads = pcb_pads;
	}else{
		TSKshipAttract(tskp);
	}

	switch(tskp->step){
		case	0:
			tskp->tskid |= TSKID_SHIP;
			tsk_ship = tskp;
			tskp->int_adr = NULL;
			tskp->exit_adr = &TSKshipExit;
			tskp->draw_adr = NULL;
			tskp->col[X] = fild(4);
			tskp->col[Y] = fild(4);
			tskp->vel[X] = fild(2);
			tskp->vel[Y] = fild(2);
			setGRPlineObj(tskp, vertex, ARRAY_SIZE(vertex));
			TSKpalleteSet(tskp, 255, 255, 255, 255);
			extend_step = 0;
			dead_flag = 0;
			limit_time = STG_LIMIT_TIME;
			break;
		case	1:
			tskp->draw_adr = &TSKshipDraw;
			SDL_PCB_SndPlaySe(SND_SHIP);
			tskp->pos[X] = fld(SCREEN_X / 2);
			tskp->pos[Y] = fld(SCREEN_Y / 2);
			tskp->acc[X] = 0;
			tskp->acc[Y] = 0;
			tskp->wait = 60;
			tskp->trg_tsk = NULL;
			tskp->old_tsk = NULL;
			ship_dirm = PAD_UP;
			tskp->roll = ship_vel[ship_dirm][Z];
			shot_wait = 0;
			tskp->step++;
		case	2:
			TSKshipTimeCtrl(tskp);
			if(tskp->int_adr == NULL){
				TSKpalleteSet(tskp, (rand() % 255), (rand() % 255), (rand() % 255), 255);
				if(tskp->wait){
					tskp->wait--;
				}else{
					TSKpalleteSet(tskp, 255, 255, 255, 255);
					tskp->int_adr = &TSKshipInt;
				}
			}
			if(ship_pads & PAD_B){
				fidiv_(tskp->acc[X], 2);
				fidiv_(tskp->acc[Y], 2);
			}
			TSKshipAccCtrl(tskp);
			if((ship_pads & (PAD_UP|PAD_DN|PAD_LF|PAD_RI))){
				fadd_(tskp->pos[X], ship_vel[ship_dirm][X]);
				fadd_(tskp->pos[Y], ship_vel[ship_dirm][Y]);
				tskp->roll = ship_vel[ship_dirm][Z];
			}
			fadd_(tskp->pos[X], tskp->acc[X]);
			fadd_(tskp->pos[Y], tskp->acc[Y]);
			TSKareaClamp(tskp);
			TSKshipShotCtrl(tskp);
			break;
		case	3:
			if(tskp->wait){
				tskp->wait--;
			}else{
				limit_time = STG_LIMIT_TIME;
				tskp->draw_adr = NULL;
				dead_flag = 0;
				tskp->step++;
			}
			tskp->int_adr = NULL;
			tskp->trg_tsk = NULL;
			tskp->old_tsk = NULL;
			fadd_(tskp->pos[X], ship_vel[ship_dirm][X]);
			fadd_(tskp->pos[Y], ship_vel[ship_dirm][Y]);
			tskp->roll = ship_vel[ship_dirm][Z];
			fadd_(tskp->pos[X], tskp->acc[X]);
			fadd_(tskp->pos[Y], tskp->acc[Y]);
			TSKareaClamp(tskp);
			break;
		case	4:
			break;
		case	10:
			TSKshipTimeCtrl(tskp);
			if(tskp->acc[X] || tskp->acc[Y]){
				tskp->roll = ship_vel[ship_dirm][Z];
				fadd_(tskp->pos[X], tskp->acc[X]);
				fadd_(tskp->pos[Y], tskp->acc[Y]);
				if(ficom(tskp->acc[X], >, 0)) fsub_(tskp->acc[X], ACC_RATE);
				if(ficom(tskp->acc[X], <, 0)) fadd_(tskp->acc[X], ACC_RATE);
				if(ficom(tskp->acc[Y], >, 0)) fsub_(tskp->acc[Y], ACC_RATE);
				if(ficom(tskp->acc[Y], <, 0)) fadd_(tskp->acc[Y], ACC_RATE);
				if(fcom(ffabs(tskp->acc[X]), <=, ACC_RATE)) tskp->acc[X] = 0;
				if(fcom(ffabs(tskp->acc[Y]), <=, ACC_RATE)) tskp->acc[Y] = 0;
				TSKareaClamp(tskp);
			}else{
				tskp->trg_tsk = NULL;
				tskp->old_tsk = NULL;
				tskp->step = 2;
			}
			TSKshipShotCtrl(tskp);
			break;
		case	20:
			if(tskp->tskid & TSKID_DESTROY){
				tskp->draw_adr = NULL;
				SDL_PCB_SndPlaySe(SND_SDEST);
				TSKpalleteSet(tskp, 255, 255, 255, 255);
				setTSKbroken(tskp, 8, fild(4));
				tskp->tskid &= ~TSKID_DESTROY;
				dead_flag = 1;
			}
			if(tskp->wait){
				tskp->wait--;
			}else{
				if(!time_off){
					limit_time += 60 * 10;
				}
				tskp->step = 1;
			}
			break;
		case	100:
			SDL_PCB_SndPlaySe(SND_SDEST);
			tskp->draw_adr = NULL;
			TSKpalleteSet(tskp, 255, 255, 255, 255);
			setTSKbroken(tskp, 8, fild(4));
			game_mode = 2;
			tskp->step++;
		case	101:
			break;

		default:
			TSKclear(tskp);
			break;
	}

	TSKshipExtend(tskp);
}

static void TSKshipInt(TSK *tskp)
{
	if(tskp->tsk_collision->tskid & TSKID_ENEMY){
		if(tskp->old_tsk == tskp->tsk_collision) return;
		tskp->old_tsk = tskp->tsk_collision;
		if(ship_dirm & PAD_RI){
			ship_dirm = PAD_LF;
			tskp->acc[X] = fimul(ship_vel[ship_dirm][X], 2);
			tskp->acc[Y] = 0;
		}else if(ship_dirm & PAD_LF){
			ship_dirm = PAD_RI;
			tskp->acc[X] = fimul(ship_vel[ship_dirm][X], 2);
			tskp->acc[Y] = 0;
		}else if(ship_dirm & PAD_DN){
			ship_dirm = PAD_UP;
			tskp->acc[X] = 0;
			tskp->acc[Y] = fimul(ship_vel[ship_dirm][Y], 2);
		}else if(ship_dirm & PAD_UP){
			ship_dirm = PAD_DN;
			tskp->acc[X] = 0;
			tskp->acc[Y] = fimul(ship_vel[ship_dirm][Y], 2);
		}
		SDL_PCB_SndPlaySe(SND_REFLECT);
		tskp->step = 10;
		return;
	}
	ship_destroy(tskp);
}

static void TSKshipExit(TSK *tskp)
{
}

static void TSKshipDraw(TSK *tskp)
{
	drawGRPlineObj(tskp);
}

static void TSKshipAccCtrl(TSK *tskp)
{
	int dir;

	dir = (ship_pads & (PAD_UP|PAD_DN|PAD_LF|PAD_RI));
	if(dir){
		if(dir & PAD_RI)		dir = PAD_RI;
		else if(dir & PAD_LF)	dir = PAD_LF;
		else if(dir & PAD_DN)	dir = PAD_DN;
		else if(dir & PAD_UP)	dir = PAD_UP;

		if(dir & PAD_RI){
			if(ship_dirm & PAD_RI){
				fadd_(tskp->acc[X], fidiv(ACC_RATE, 2));
				if(fcom(tskp->acc[X], >, ACC_MAX)) tskp->acc[X] = ACC_MAX;
			}else if(ship_dirm & PAD_LF){
				tskp->acc[X] = 0;
			}else if(ship_dirm & PAD_DN){
				tskp->acc[X] = ffabs(tskp->acc[Y]);
				tskp->acc[Y] = 0;
			}else if(ship_dirm & PAD_UP){
				tskp->acc[X] = ffabs(tskp->acc[Y]);
				tskp->acc[Y] = 0;
			}
		}else if(dir & PAD_LF){
			if(ship_dirm & PAD_LF){
				fsub_(tskp->acc[X], fidiv(ACC_RATE, 2));
				if(fcom(tskp->acc[X], <, fchs(ACC_MAX))) tskp->acc[X] = fchs(ACC_MAX);
			}else if(ship_dirm & PAD_RI){
				tskp->acc[X] = 0;
			}else if(ship_dirm & PAD_DN){
				tskp->acc[X] = fchs(ffabs(tskp->acc[Y]));
				tskp->acc[Y] = 0;
			}else if(ship_dirm & PAD_UP){
				tskp->acc[X] = fchs(ffabs(tskp->acc[Y]));
				tskp->acc[Y] = 0;
			}
		}else if(dir & PAD_DN){
			if(ship_dirm & PAD_DN){
				fadd_(tskp->acc[Y], fidiv(ACC_RATE, 2));
				if(fcom(tskp->acc[Y], >, ACC_MAX)) tskp->acc[Y] = ACC_MAX;
			}else if(ship_dirm & PAD_UP){
				tskp->acc[Y] = 0;
			}else if(ship_dirm & PAD_RI){
				tskp->acc[Y] = ffabs(tskp->acc[X]);
				tskp->acc[X] = 0;
			}else if(ship_dirm & PAD_LF){
				tskp->acc[Y] = ffabs(tskp->acc[X]);
				tskp->acc[X] = 0;
			}
		}else if(dir & PAD_UP){
			if(ship_dirm & PAD_UP){
				fsub_(tskp->acc[Y], fidiv(ACC_RATE, 2));
				if(fcom(tskp->acc[Y], <, fchs(ACC_MAX))) tskp->acc[Y] = fchs(ACC_MAX);
			}else if(ship_dirm & PAD_DN){
				tskp->acc[Y] = 0;
			}else if(ship_dirm & PAD_RI){
				tskp->acc[Y] = fchs(ffabs(tskp->acc[X]));
				tskp->acc[X] = 0;
			}else if(ship_dirm & PAD_LF){
				tskp->acc[Y] = fchs(ffabs(tskp->acc[X]));
				tskp->acc[X] = 0;
			}
		}
		ship_dirm = dir;
	}else{
		if(fcom(tskp->acc[X], >, 0)) fsub_(tskp->acc[X], ACC_RATE);
		if(fcom(tskp->acc[X], <, 0)) fadd_(tskp->acc[X], ACC_RATE);
		if(fcom(tskp->acc[Y], >, 0)) fsub_(tskp->acc[Y], ACC_RATE);
		if(fcom(tskp->acc[Y], <, 0)) fadd_(tskp->acc[Y], ACC_RATE);
		if(fcom(ffabs(tskp->acc[X]), <=, ACC_RATE)) tskp->acc[X] = 0;
		if(fcom(ffabs(tskp->acc[Y]), <=, ACC_RATE)) tskp->acc[Y] = 0;
	}
}

static void TSKshipShotCtrl(TSK *tskp)
{
	TSK *etskp;
	fixed acc = ffabs(fadd(tskp->acc[X], tskp->acc[Y]));

	if(!shot_wait){
		if(ship_pads & PAD_A){
			etskp = TSKsetBtm(&TSKsshot, GROUP_SSHOT);
			if(etskp){
				etskp->roll = tskp->roll;
				etskp->len = fadd(SSHOT_LENGTH, fimul(acc, 8));
				etskp->level = fist(fdiv(acc, (fidiv(fisub(ACC_MAX, 1), 4))));
				if(ship_dirm & PAD_LF){
					etskp->pos[X] = fisub(tskp->pos[X], 8);
					etskp->pos[Y] =       tskp->pos[Y];
					etskp->vel[X] = fsub(fmul(tskp->acc[X], fld(1.5)), SSHOT_SPD);
					etskp->vel[Y] = 0;
				}else if(ship_dirm & PAD_RI){
					etskp->pos[X] = fiadd(tskp->pos[X], 8);
					etskp->pos[Y] = tskp->pos[Y];
					etskp->vel[X] = fadd(fmul(tskp->acc[X], fld(1.5)), SSHOT_SPD);
					etskp->vel[Y] = 0;
				}else if(ship_dirm & PAD_UP){
					etskp->pos[X] = tskp->pos[X];
					etskp->pos[Y] = fisub(tskp->pos[Y], 8);
					etskp->vel[X] = 0;
					etskp->vel[Y] = fsub(fmul(tskp->acc[Y], fld(1.5)), SSHOT_SPD);
				}else if(ship_dirm & PAD_DN){
					etskp->pos[X] = tskp->pos[X];
					etskp->pos[Y] = fiadd(tskp->pos[Y], 8);
					etskp->vel[X] = 0;
					etskp->vel[Y] = fadd(fmul(tskp->acc[Y], fld(1.5)), SSHOT_SPD);
				}
				shot_wait = 10 - fist(acc);
			}
			TSKareaClamp(etskp);
		}
	}else{
		shot_wait--;
	}
}

static void TSKshipExtend(TSK *tskp)
{
	if(tskp->step >= 100) return;
	if(attract_mode != 0) return;

	switch(extend_step){
		case	0:
			extend_score = EXTEND_1ST;
			extend_step++;
		case	1:
			if(extend_score <= score){
				if(left < 4){
					SDL_PCB_SndPlaySe(SND_EXTEND);
					left++;
				}
				extend_step++;
				extend_score += EXTEND_2ND;
			}
			break;
		case	2:
			if(extend_score <= score){
				if(left < 4){
					SDL_PCB_SndPlaySe(SND_EXTEND);
					left++;
				}
				extend_score += EXTEND_3RD;
			}
			break;
	}
}

static void TSKshipTimeCtrl(TSK *tskp)
{
	if(!time_off){
		if(limit_time){
			limit_time--;
		}
	}
	if(limit_time && limit_time <= (60 * 11) && !(limit_time % 60)){
		SDL_PCB_SndPlaySe(SND_CNT00 + (limit_time / 60) - 1);
	}
}

static void TSKshipAttract(TSK *tskp)
{
	TSK *etskp;
	fixed dist = fild(1000);
	fixed px, py;
	fixed dx, dy;

	ship_pads = PAD_A;

	px = tskp->pos[X];
	py = tskp->pos[Y];

	if(tskp->trg_tsk == NULL){
		for(etskp = TskIndex[GROUP_ENEMY][TSK_TOP].next_tsk; etskp != &(TskIndex[GROUP_ENEMY][TSK_BOTTOM]); etskp = etskp->next_tsk){
			if(etskp->step < 100 && etskp->step != -1){
				dx = ffabs(fsub(etskp->pos[X], px));
				dy = ffabs(fsub(etskp->pos[Y], py));
				if(fcom(dist, >, fadd(dx, dy))) {
					tskp->trg_tsk = etskp;
					dist = fadd(dx, dy);
				}
			}
		}
	}

	if(tskp->trg_tsk == NULL) return;

	dx = tskp->trg_tsk->pos[X];
	dy = tskp->trg_tsk->pos[Y];

	if(ficom(ffabs(fsub(px, dx)), >, 4)){
		if(fcom(px, <, dx)) ship_pads |= PAD_RI;
		else				ship_pads |= PAD_LF;
	}else if(ficom(ffabs(fsub(py, dy)), >, 4)){
		if(fcom(py, <, dy)) ship_pads |= PAD_DN;
		else				ship_pads |= PAD_UP;
	}else{
		if(fcom(ffabs(fsub(px, dx)), >, ffabs(fsub(py, dy)))){
			if(fcom(px, <, dx)) ship_pads |= PAD_RI;
			else				ship_pads |= PAD_LF;
		}else{
			if(fcom(py, <, dy)) ship_pads |= PAD_DN;
			else				ship_pads |= PAD_UP;
		}
	}
}

static void ship_destroy(TSK *tskp)
{
	tskp->tskid |= TSKID_DESTROY;
	tskp->int_adr = NULL;
	if(left){
		if(!attract_mode) left--;
		tskp->wait = 60;
		tskp->step = 20;
	}else{
		tskp->step = 100;
	}
}
